from typing import Any
from .dao import Dao

"""
Service that allows to manage jobs
"""


class JobService:
    def __init__(self, maximum_runtime: float = 3000):
        self.dao = Dao.get_instance()

    def get_job_result(self, job_id: str | None) -> dict[str, Any] | None:
        if job_id is None:
            return None
        return self.dao.get_job_result(job_id)

    def post_job_result(self, job_id: str, result: Any):
        self.dao.post_job_result(job_id, result)

    def job_exists(self, job_id: str) -> bool:
        return self.dao.job_exists(job_id)

    def delete_job(self, job_id: str):
        self.dao.delete_job(job_id)

    def reset(self):
        self.dao.clear_job_results()
        self.dao.clear_small_memory()



from .flojoy_instruction import FLOJOY_INSTRUCTION
from .plotly_utils import data_container_to_plotly
from .data_container import DataContainer
from .dao import Dao
from typing import Any, cast, Optional

__all__ = ["get_job_result", "get_next_directions", "get_next_nodes", "get_job_result"]


def is_flow_controled(result: dict[str, Any] | DataContainer):
    if (
        FLOJOY_INSTRUCTION.FLOW_TO_DIRECTIONS in result
        or FLOJOY_INSTRUCTION.FLOW_TO_NODES in result
    ):
        return True
    return False


def get_next_directions(result: dict[str, Any] | None) -> list[str] | None:
    direction = None
    if result is None:
        return direction
    if not result.get(FLOJOY_INSTRUCTION.FLOW_TO_DIRECTIONS):
        for value in result.values():
            if isinstance(value, dict) and value.get(
                FLOJOY_INSTRUCTION.FLOW_TO_DIRECTIONS
            ):
                direction = cast(
                    list[str], value[FLOJOY_INSTRUCTION.FLOW_TO_DIRECTIONS]
                )
                break
    else:
        direction = result[FLOJOY_INSTRUCTION.FLOW_TO_DIRECTIONS]
    return direction


def get_next_nodes(result: dict[str, Any] | None) -> list[str]:
    if result is None:
        return []
    return cast(list[str], result.get(FLOJOY_INSTRUCTION.FLOW_TO_NODES, []))


def get_dc_from_result(
    result: dict[str, Any] | DataContainer | None
) -> DataContainer | None:
    if not result:
        return None
    if isinstance(result, DataContainer):
        return result
    if result.get(FLOJOY_INSTRUCTION.RESULT_FIELD):
        return result[result[FLOJOY_INSTRUCTION.RESULT_FIELD]]
    return result["data"]


def get_job_result(job_id: str) -> dict[str, Any] | DataContainer | None:
    try:
        job_result: Any = Dao.get_instance().get_job_result(job_id)
        result = get_dc_from_result(cast(dict[str, Any] | DataContainer, job_result))
        return result
    except Exception:
        return None


def get_text_blob_from_dc(dc: DataContainer) -> str | None:
    match dc.type:
        case "TextBlob":
            return dc.text_blob
        case "Bytes":
            return dc.b.decode("utf-8")
        case _:
            return None


def get_frontend_res_obj_from_result(
    result: Optional[dict[str, Any] | DataContainer]
) -> Optional[dict[str, Any]]:
    if result is None:
        return None

    if isinstance(result, DataContainer):
        plotly_fig = data_container_to_plotly(data=result)
        return {"plotly_fig": plotly_fig, "text_blob": get_text_blob_from_dc(result)}
    if result.get(FLOJOY_INSTRUCTION.RESULT_FIELD):
        data = result[result[FLOJOY_INSTRUCTION.RESULT_FIELD]]
        if not data:
            return result

        plotly_fig = None
        if isinstance(data, DataContainer):
            plotly_fig = data_container_to_plotly(data=data)
        return {
            **result,
            "plotly_fig": plotly_fig,
            "text_blob": get_text_blob_from_dc(data),
        }
    keys = list(result.keys())
    return get_frontend_res_obj_from_result(result[keys[0]])



import numpy as np
import pandas as pd
from typing import Any, Callable
from threading import Lock
from .data_container import DCNpArrayType

MAX_LIST_SIZE = 1000

_dict_sm_lock = Lock()  # dict small memory lock
_dict_job_lock = Lock()  # dict job lock
_dict_node_init_container_lock = Lock()
_dict_node_init_func_lock = Lock()
_init_lock = Lock()

"""
Used by clients to create a new instance of the datastorage
"""


def create_storage():
    return Dao.get_instance()


"""
This class is a Singleton that acts as a in-memory datastorage

IMPORTANT: The commented code should not be removed, as it acts as a reference for the future
in case we need to implement a Redis based datastorage
"""


class Dao:
    _instance = None

    @classmethod
    def get_instance(cls):
        with _init_lock:
            if Dao._instance is None:
                Dao._instance = Dao()
            return Dao._instance

    def __init__(self):
        self.storage = {}  # small memory
        self.job_results = {}
        self.node_init_container = {}
        self.node_init_func = {}

    """
    METHODS FOR JOB RESULTS
    """

    def get_job_result(self, job_id: str) -> Any | None:
        with _dict_job_lock:
            res = self.job_results.get(job_id, None)
        if res is None:
            raise ValueError(f"Job result with id {job_id} does not exist")
        return res

    def post_job_result(self, job_id: str, result: Any):
        with _dict_job_lock:
            self.job_results[job_id] = result

    def clear_job_results(self):
        with _dict_job_lock:
            self.job_results.clear()

    def job_exists(self, job_id: str) -> bool:
        with _dict_job_lock:
            return job_id in self.job_results.keys()

    def delete_job(self, job_id: str):
        with _dict_job_lock:
            self.job_results.pop(job_id, None)

    """
    METHODS FOR SMALL MEMORY
    """

    def clear_small_memory(self):
        with _dict_sm_lock:
            self.storage.clear()

    def check_if_valid(self, result: Any | None, expected_type: Any):
        with _dict_sm_lock:
            if result is not None and not isinstance(result, expected_type):
                raise ValueError(
                    f"Expected {expected_type} type, but got {type(result)} instead!"
                )

    def set_np_array(self, memo_key: str, value: DCNpArrayType):
        with _dict_sm_lock:
            self.storage[memo_key] = value

    def set_pandas_dataframe(self, key: str, dframe: pd.DataFrame):
        with _dict_sm_lock:
            self.storage[key] = dframe

    def set_str(self, key: str, value: str):
        with _dict_sm_lock:
            self.storage[key] = value

    def get_pd_dataframe(self, key: str) -> pd.DataFrame | None:
        with _dict_sm_lock:
            encoded = self.storage.get(key, None)
        self.check_if_valid(encoded, pd.DataFrame)
        return encoded

    def get_np_array(self, memo_key: str) -> DCNpArrayType | None:
        with _dict_sm_lock:
            encoded = self.storage.get(memo_key, None)
        self.check_if_valid(encoded, np.ndarray)
        return encoded

    def get_str(self, key: str) -> str | None:
        with _dict_sm_lock:
            encoded = self.storage.get(key, None)
        return encoded

    def get_obj(self, key: str) -> dict[str, Any] | None:
        with _dict_sm_lock:
            r_obj = self.storage.get(key, None)
        self.check_if_valid(r_obj, dict)
        return r_obj

    def set_obj(self, key: str, value: dict[str, Any]):
        with _dict_sm_lock:
            self.storage[key] = value

    def delete_object(self, key: str):
        with _dict_sm_lock:
            self.storage.pop(key)

    def remove_item_from_set(self, key: str, item: Any):
        with _dict_sm_lock:
            res = self.storage.get(key, None)
        self.check_if_valid(res, set)
        if not res:
            return
        with _dict_sm_lock:
            res.remove(item)

    def add_to_set(self, key: str, value: Any):
        with _dict_sm_lock:
            res: set[Any] | None = self.storage.get(key, None)
        if res is None:
            res = set()
            res.add(value)
            self.storage[key] = res
            return
        self.check_if_valid(res, set)
        with _dict_sm_lock:
            res.add(value)

    def get_set_list(self, key: str) -> list[Any] | None:
        with _dict_sm_lock:
            res = self.storage.get(key, None)
        if res is None:
            return None
        self.check_if_valid(res, set)
        return list(res)

    """
    METHODS FOR NODE INIT
    """

    # -- for node container --
    def clear_node_init_containers(self):
        with _dict_node_init_container_lock:
            self.node_init_container.clear()

    def set_init_container(self, node_id: str, value):
        with _dict_node_init_container_lock:
            self.node_init_container[node_id] = value

    def get_init_container(self, node_id: str):
        with _dict_node_init_container_lock:
            res = self.node_init_container.get(node_id, None)
        from .node_init import NodeInitContainer  # avoid circular import

        self.check_if_valid(res, NodeInitContainer)
        return res

    def has_init_container(self, node_id: str) -> bool:
        with _dict_node_init_container_lock:
            return node_id in self.node_init_container.keys()

    # ------------------------

    # -- for node init function --
    def set_init_function(self, node_func, node_init_func):
        with _dict_node_init_func_lock:
            self.node_init_func[node_func] = node_init_func

    def get_init_function(self, node_func: Callable):
        with _dict_node_init_func_lock:
            res = self.node_init_func.get(node_func, None)
        from .node_init import NodeInit  # avoid circular import

        self.check_if_valid(res, NodeInit)
        return res

    def has_init_function(self, node_func) -> bool:
        with _dict_node_init_func_lock:
            return node_func in self.node_init_func.keys()

    # ----------------------------



from .data_container import *
from .flojoy_python import *
from .job_result_builder import *
from .flojoy_instruction import *
from .plotly_utils import *
from .module_scraper import *
from .job_result_utils import *
from .data_container import *
from .utils import *
from .parameter_types import *
from .small_memory import *
from .flojoy_node_venv import *
from .job_service import *
from .node_init import *
from .config import *
from .node_preflight import *
from .flojoy_cloud import *
